﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;
using System.Xml;

namespace RewriteExpressionTree
{
    class Program
    {
        public static int? Add(int? x, int? y)
        {
            return x + y;
        }
        public class Node
        {
            public Node Next;
            public string Value;
        }
        static void Main(string[] args)
        {
            var get0 = LiftMemberAccessToNull((Node f) => f.Value);
            var get1 = LiftMemberAccessToNull((Node f) => f.Next.Value);
            var get2 = LiftMemberAccessToNull((Node f) => f.Next.Next.Value);
            var get3 = LiftMemberAccessToNull((Node f) => f.Next.Next.Next.Value);

            Node f1 = new Node { Value = "One" };
            Node f2 = new Node { Value = "Two" };
            Node f3 = new Node { Value = "Three" };
            f1.Next = f2; f2.Next = f3;

            Node[] foos = { f1, f2, f3, null };
            foreach (Node f in foos)
            {
                Console.WriteLine("Value: " + (get0(f) ?? "null"));
                Console.WriteLine("Next.Value: " + (get1(f) ?? "null"));
                Console.WriteLine("Next.Next.Value: " + (get2(f) ?? "null"));
                Console.WriteLine("Next.Next.Next.Value: " + (get3(f) ?? "null"));

                Console.WriteLine();
            }
        }
        


        /// <summary>
        /// 
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <typeparam name="TResult"></typeparam>
        /// <param name="expr"></param>
        /// <returns></returns>
        /// <remarks>
        /// What happens if this encounters a Nullable?
        /// </remarks>
        static Func<T, TResult> LiftMemberAccessToNull<T, TResult>(Expression<Func<T, TResult>> expr)
        {
            NullLiftModifier modifier = new NullLiftModifier();
            var x = (Expression<Func<T, TResult>>) modifier.Modify(expr);
            return x.Compile();
        }
    }

    public class NullLiftModifier : ExpressionVisitor
    {
        // The method that kicks off the tree walk is
        // protected, so we need to define a public
        // method that provides access to it.
        public Expression Modify(Expression expression)
        {
            return Visit(expression);
        }

        protected override Expression VisitMemberAccess(MemberExpression originalExpression)
        {
            MemberExpression memberAccessExpression = (MemberExpression) base.VisitMemberAccess(originalExpression);

            // This is a bit messy. We end up evaluating the expression
            // that retrieves the object whose member we're like to access
            // twice! Once to test it for null, and if it's not, we evaluate
            // it a second time in order to perform the member access.
            // What we want is a 'let' construct, but expression trees
            // don't appear to have such a thing. (LINQ query expressions
            // offer 'let', but I believe C# ends up building it out of a lambda.)
            Expression nullTest = Expression.Equal(memberAccessExpression.Expression, Expression.Constant(null, memberAccessExpression.Expression.Type));
            return Expression.Condition(nullTest, Expression.Constant(null, memberAccessExpression.Type), memberAccessExpression);
        }
    }
}
